/*-
 * Copyright (c) 2008 Jean-Yves Moulin <jym@ldh.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <netdb.h>

#include "lbdr.h"
#include "common.h"


int debug_ioctl = 0;
int paddr_resolv = 1;
int hformat = 1;
int af = PF_UNSPEC;
int no_modify = 0;


/**
 * Use only one time in a printf()
 * because of local addr (=> no malloc)
 */
#define IP_BUF (NI_MAXHOST + NI_MAXSERV)

char ip_buf[IP_BUF];

char *paddr(struct sockaddr_storage *ip)
{
	int flags = NI_NOFQDN|NI_NUMERICSERV;
	char name_buf[NI_MAXHOST];
	char serv_buf[NI_MAXSERV];
	char *c = NULL;
	int err = 0;

	bzero(ip_buf, IP_BUF);
	bzero(name_buf, NI_MAXHOST);
	bzero(serv_buf, NI_MAXSERV);

	if(!paddr_resolv) flags |= NI_NUMERICHOST;

	err = getnameinfo((struct sockaddr *)ip, ip->ss_len, name_buf, sizeof(name_buf), NULL, 0, NI_NAMEREQD|flags);

	getnameinfo((struct sockaddr *)ip, ip->ss_len, name_buf, sizeof(name_buf), serv_buf, sizeof(serv_buf), flags);

	c = strchr(name_buf, '.');
	if(!err && c && paddr_resolv)
		*c = '\0';

	snprintf(ip_buf, IP_BUF, "%s%c%s", name_buf, ip->ss_family == AF_INET ? ':' : '.', serv_buf);

	return ip_buf;
}


int same_addr(struct sockaddr_storage *a, struct sockaddr_storage *b)
{
	struct sockaddr_in *in4a, *in4b;
	struct sockaddr_in6 *in6a, *in6b;

	if(a->ss_family != b->ss_family) return 0;

        switch (a->ss_family) {
        case AF_INET:
		in4a = (struct sockaddr_in *)a;
		in4b = (struct sockaddr_in *)b;
                if( in4a->sin_addr.s_addr == in4b->sin_addr.s_addr &&
                           in4a->sin_port == in4b->sin_port)
			 return 1;
                break;
        case AF_INET6:
        	in6a = (struct sockaddr_in6 *)a;
        	in6b = (struct sockaddr_in6 *)b;

                if( in6a->sin6_addr.s6_addr[0]  == in6b->sin6_addr.s6_addr[0] &&
                    in6a->sin6_addr.s6_addr[4]  == in6b->sin6_addr.s6_addr[4] &&
                    in6a->sin6_addr.s6_addr[8]  == in6b->sin6_addr.s6_addr[8] &&
                    in6a->sin6_addr.s6_addr[12] == in6b->sin6_addr.s6_addr[12] &&
		in6a->sin6_port == in6b->sin6_port) 
			return 1;
                break;
        }

	return 0;
}

int sched_id(char *s)
{
	if(!strncasecmp("wlc", s, 3)) return LBDR_SCHED_WLC;
	else if(!strncasecmp("wrr", s, 3)) return LBDR_SCHED_WRR;
	else if(!strncasecmp("lc", s, 2)) return LBDR_SCHED_LC;
	else if(!strncasecmp("rr", s, 2)) return LBDR_SCHED_RR;

	return 0;
}


char *sched_name(int type, int full)
{
	if(full) {
		switch(type) {
			case LBDR_SCHED_WLC:     return "Weighted Least-Connection";
			case LBDR_SCHED_WRR:     return "Weighted Round-Robin";
			case LBDR_SCHED_LC:      return "Least-Connection";
			case LBDR_SCHED_RR:      return "Round-Robin";
			default: break;
		}
	} else {
		switch(type) {
			case LBDR_SCHED_WLC: return "wlc";
			case LBDR_SCHED_WRR: return "wrr";
			case LBDR_SCHED_LC: return "lc";
			case LBDR_SCHED_RR: return "rr";
			default: break;
		}
	}

	return "***";
}

char *ioctl_type(long c)
{
	switch(c) {
		case LBDRCLEAR: return "Clear-Rules";
		case LBDRADDCONF: return "Add-Conf";
		case LBDRDELCONF: return "Del-Conf";
		case LBDRCHGCONF: return "Change-Conf";
		case LBDRCLONE: return "Clone-Service";
		case LBDRINFO: return "Get-Info";
		case LBDRSTATS: return "Get-Stats";
		default: break;
	}

	return "(unknown)";
}



int is_digit(char *s)
{
	int i;
	for(i = 0; s[i]; i++) {
		if(s[i] != '0' &&
		   s[i] != '1' &&
		   s[i] != '2' &&
		   s[i] != '3' &&
		   s[i] != '4' &&
		   s[i] != '5' &&
		   s[i] != '6' &&
		   s[i] != '7' &&
		   s[i] != '8' &&
		   s[i] != '9')
		   	return 0;
	}

	return 1;
}



/**
 * My own print_human() function.
 * The netbsd and freebsd is not as I want :)
 * BE CAREFUL: USE IT ONLY ONE TIME IN PRINTF() 
 */
#define HU_MAXBUF 32

char buffer[HU_MAXBUF];

char *human(double nb, char *suffix, int flags)
{
	int div = 0;
	int mul = 1024;
	const char *prefixes = "\0\0K\0M\0G\0T\0P\0E";

	if(flags & HU_DIVISOR_1000) mul = 1000;

	bzero(buffer, HU_MAXBUF);

	if(hformat) {
		while(nb >= 1000) {
			div++;
			nb = nb / mul;
		}

		if(/*(flags & HU_DECIMAL) &&*/ nb < 10)
			snprintf(buffer, HU_MAXBUF, "%.1f%s%s%s", nb, flags & HU_NOSPACE ? "" : " ", &prefixes[div << 1], suffix);
		else
			snprintf(buffer, HU_MAXBUF, "%.0f%s%s%s", nb, flags & HU_NOSPACE ? "" : " ", &prefixes[div << 1], suffix);
	} else {
		if(/*(flags & HU_DECIMAL) ||*/ nb < 10)
			snprintf(buffer, HU_MAXBUF, "%.1f%s", nb, suffix);
		else
			snprintf(buffer, HU_MAXBUF, "%.0f%s", nb, suffix);
	}

	return buffer;
}



/**
 * send ioctl to drlb module
 */
int lbdr_ioctl(long cmd, void *arg)
{
	int fd, r;

	if(no_modify) {
		return 0;
	}

	fd = open(LBDR_DEVICE, O_RDWR);

	if(fd == -1) {
		if(errno == ENOENT) {
			fprintf(stderr, "Unable to open drlb device. Is module loaded ?");
		}
		if(!no_modify) {
			perror("open drlb");
		}
		return -1;
	}

	r = ioctl(fd, cmd, arg);

	close(fd);

	return r;
}

/**
 * Function used to send configuration to drlb (change server or service parameters..Etc)
 * IP input are in strings
 */ 
int lbdr_change_server(long cmd, struct sockaddr_storage *ipdst, struct sockaddr_storage *ipsrv, int hashtablesize, int sched, int status, int weight, int rate, int tls, int options, int t_syn, int t_est, int t_clo)
{
	struct lbdr_conf cnf;

	bcopy(ipdst, &(cnf.ip_dst), sizeof(struct sockaddr_storage));

	if(ipsrv) {
		bcopy(ipsrv, &(cnf.ip_srv), sizeof(struct sockaddr_storage));
		if(cnf.ip_srv.ss_family != cnf.ip_dst.ss_family) {
			printf("address family must be the same (%d != %d)\n",
			cnf.ip_dst.ss_family, cnf.ip_srv.ss_family);
			return -1;
		}

		if((SPORT(&(cnf.ip_srv))) != (SPORT(&(cnf.ip_dst)))) {
			printf("port must be the same (%d != %d)\n",
				ntohs(SPORT(&(cnf.ip_srv))), ntohs(SPORT(&(cnf.ip_dst))));
			return -1;
		}
	}

	cnf.t_syn = t_syn;
	cnf.t_est = t_est;
	cnf.t_clo = t_clo;
	cnf.sched = sched;
	cnf.persist = tls;
	cnf.weight = weight;
	cnf.hashtablesize = hashtablesize;
	cnf.rate = rate;
	cnf.status = status;
	cnf.options = options;
	cnf.force = 0;
	cnf.clone = 0;

	if(debug_ioctl) {
		printf("%s '%s' IOCTL with parameter:\n"
			"\tService: %s\n"
			 , no_modify ? "Simulate" : "Sending"
			 , ioctl_type(cmd), paddr(&(cnf.ip_dst)));
		printf( "\tServer: %s\n"
			"\tClone service: %s\n"
			"\tScheduler: %s\n"
			"\tHashtableSize: 2^%d\n"
			"\tPersistence: %s\n"
			"\tTimeout SYN: %d\n"
			"\tTimeout ESTABLISH: %d\n"
			"\tTimeout CLOSED: %d\n"
			"\tServer Weight: %d\n"
			"\tServer Max cnx/s: %d\n"
			"\tServer Status: %s\n"
			"\tForce action: %s\n"
			"\tOptions field: 0x%x\n\n"
			, paddr(&(cnf.ip_srv)), cnf.clone ? "True" : "False",
			  sched_name(cnf.sched, 1), cnf.hashtablesize,
			  cnf.persist ? "True" : "False",
			  cnf.t_syn, cnf.t_est, cnf.t_clo,
			  cnf.weight, cnf.rate, cnf.status ? "Active" : "INactive",
			  cnf.force ? "True" : "False", options);
	}


	return lbdr_ioctl(cmd, &cnf);
}


int lbdr_set_brother(char *addr, char *key)
{
	struct lbdr_share shr;
	struct sockaddr_storage *ip;

	ip = get_ip(addr, "10149", af);

	if(!ip) {
		puts("Error in address resolution.");
		return -1;
	}

	memcpy(&(shr.ip_dst), ip, ip->ss_len);
	free(ip);

	if(sscanf(key, "%i", &shr.key) != 1) {
		puts("You must provide a key (in hexa or decimal).");
		return -1;
	}

	if(debug_ioctl)
		printf( "%s 'Set-Brother' IOCTL with parameter:\n"
			"\tBrother IP: %s\n"
			"\tBrother KEY: 0x%x\n"
			 , no_modify ? "Simulate" : "Sending"
			, paddr(&(shr.ip_dst)), shr.key);

	return lbdr_ioctl(LBDRSETSHARE, &shr);
}

/** 
 * Want to know how many IP are in drlb (server + service + clone)
 * Can output some other info (malloc error, badmbuf, badudp...)
 */
int lbdr_get_nip(int print)
{
	struct lbdr_info info;

	bzero(&info, sizeof(struct lbdr_info));

	if(debug_ioctl) printf("%s 'Get-Info' IOCTL\n"
			 , no_modify ? "Simulate" : "Sending" );

	lbdr_ioctl(LBDRINFO, &info);

	if(debug_ioctl) printf("Number of registred IP in drlb: %d\n", info.registred_ip);

	if(print) {
		printf( "Errors:\n"
			"\tNo memory available: %ld\n"
			"\tDropped active cnx: %ld\n"
			"\tmbuf too short: %ld\n"
			"\tCan't allocate mbuf: %ld\n"
			"\tUDP bad checksum: %ld\n"
			"\tUDP bad lenght: %ld\n"
			"\tUnknow shared server: %ld\n"
			, info.err_nomem, info.err_dropact
			, info.err_badmbuf, info.err_nombuf
			, info.err_udpbadcsum, info.err_udpbadlen
			, info.err_notsame);

		if(info.brother_ip.ss_family) {
			printf("Sharing with %s Key: 0x%x\n\n",
				paddr(&(info.brother_ip)), info.brother_key);
		}
	}

	return info.registred_ip;
}

/**
 * Fetch all statistic for all IP from drlb module
 */
struct lbdr_param * lbdr_get_stats(int n)
{
	struct lbdr_param *prm;

	prm = malloc(sizeof(struct lbdr_param));
	if(!prm) {
		puts("Can't allocate memory!");
		exit(EXIT_FAILURE);
	}

	prm->size = n * sizeof(struct lbdr_stats);
	prm->data = malloc(prm->size);
	if(!prm->data) {
		puts("Can't allocate memory!");
		exit(EXIT_FAILURE);
	}
	bzero(prm->data, prm->size);

	if(debug_ioctl) printf("%s 'Get-Stats' IOCTL\n"
			 , no_modify ? "Simulate" : "Sending" );

	lbdr_ioctl(LBDRSTATS, prm);

	return prm;

}

/**
 * Take a string in format "127.0.0.1,80" in order to fill a struct sockaddr
 * Be careful: separator is currently ','
 */
struct sockaddr_storage * get_ip(char *av, char *defp, int family)
{
	struct addrinfo hints, *res0;
	struct sockaddr_storage *addr = NULL;
	int error;
	char *port;

	port = strchr(av, ',');
	if(port) {
		*port++ = '\0';
		if(!is_digit(port)) {
			printf("'%s' is not a valid port", port);
			port = defp;
		}
	} else {
		port = defp;
	}
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = family;
	hints.ai_socktype = SOCK_STREAM;

	error = getaddrinfo(av, port, &hints, &res0);

	if (error) {
		printf("getaddrinfo: %s", gai_strerror(error));
		return NULL;
	}

	if(res0) {
		addr = malloc(sizeof(struct sockaddr_storage));
		if(addr)
			memcpy(addr, res0->ai_addr, sizeof(struct sockaddr_storage));
		else
			perror("malloc");
	} else {
		printf("getaddrinfo: %s", gai_strerror(error));
	}

	freeaddrinfo(res0);

	if(!addr) {
		return NULL;
	}

	if(debug_ioctl) printf("Input Read IP address: %s\n", paddr(addr));

	return addr;
}

/* EOF */
